﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionaryTracking {
  /// An analyzer that looks at individual lineages in the genealogy graph
  /// </summary>
  [Item("GeneticOperatorsAverageImprovementAnalyzer", "An operator that records the average success of genetic operators such as crossover and mutation.")]
  [StorableClass]
  class GeneticOperatorsAverageImprovementAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region Parameter names
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";
    private const string PopulationGraphParameterName = "PopulationGraph";
    private const string GeneticOperatorsAverageImprovementResultName = "GeneticOperatorsAverageImprovement";
    #endregion

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    #endregion

    #region Properties
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public bool EnabledByDefault { get { return true; } }
    #endregion


    [StorableConstructor]
    private GeneticOperatorsAverageImprovementAnalyzer(bool deserializing) : base(deserializing) { }
    private GeneticOperatorsAverageImprovementAnalyzer(GeneticOperatorsAverageImprovementAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GeneticOperatorsAverageImprovementAnalyzer(this, cloner);
    }
    public GeneticOperatorsAverageImprovementAnalyzer() {
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far"));

      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }

    #region AfterDeserialization
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }
    #endregion

    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    public override IOperation Apply() {
      UpdateCounter.Value++;
      if (UpdateCounter.Value == UpdateInterval.Value) {
        UpdateCounter.Value = 0;

        if (!Results.ContainsKey(PopulationGraphParameterName)) return base.Apply();

        var graph = (SymbolicExpressionTreeGenealogyGraph)Results[PopulationGraphParameterName].Value;

        var nodes = graph.Nodes.Where(n => n.Rank == Generations.Value); // get the nodes in the last generation and their parents

        var crossoverImprovements = new List<double>();
        var mutationImprovements = new List<double>();

        foreach (var node in nodes) {
          if (node.InEdges == null) continue; // it means the corresponding individual did not reproduce
          var parentsQuality = node.InEdges.Select(e => e.Source).Max(v => ((SymbolicExpressionGenealogyGraphNode)v).Quality);
          var childQuality = node.Quality;
          var improvement = childQuality / parentsQuality;
          if (double.IsNaN(improvement) || double.IsInfinity(improvement) || parentsQuality.IsAlmost(0.0)) improvement = 0;
          switch (node.InEdges.Count) {
            case 2: // crossover
              crossoverImprovements.Add(improvement);
              break;
            case 1: // mutation
              mutationImprovements.Add(improvement);
              break;
          }
        }

        if (!Results.ContainsKey(GeneticOperatorsAverageImprovementResultName)) {
          Results.Add(new Result(GeneticOperatorsAverageImprovementResultName, new DataTable()));
        }
        var table = Results[GeneticOperatorsAverageImprovementResultName].Value as DataTable;

        double cxAvgImprovement = 0.0;
        double cxMaxImprovement = 0.0;
        if (crossoverImprovements.Count > 0) {
          cxAvgImprovement = crossoverImprovements.Average();
          cxMaxImprovement = crossoverImprovements.Max();
        }
        double mutAvgImprovement = 0.0;
        double mutMaxImprovement = 0.0;
        if (mutationImprovements.Count > 0) {
          mutAvgImprovement = mutationImprovements.Average();
          mutMaxImprovement = mutationImprovements.Max();
        }

        if (!table.Rows.ContainsKey("Average crossover improvement")) {
          table.Rows.Add(new DataRow("Average crossover improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Average crossover improvement"].Values.Add(cxAvgImprovement);

        if (!table.Rows.ContainsKey("Average mutation improvement")) {
          table.Rows.Add(new DataRow("Average mutation improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Average mutation improvement"].Values.Add(mutAvgImprovement);

        if (!table.Rows.ContainsKey("Best crossover improvement")) {
          table.Rows.Add(new DataRow("Best crossover improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Best crossover improvement"].Values.Add(cxMaxImprovement);

        if (!table.Rows.ContainsKey("Best mutation improvement")) {
          table.Rows.Add(new DataRow("Best mutation improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Best mutation improvement"].Values.Add(mutMaxImprovement);

        if (!table.Rows.ContainsKey("Average improvement")) {
          table.Rows.Add(new DataRow("Average improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Average improvement"].Values.Add((cxAvgImprovement + mutAvgImprovement) / 2);

        if (!table.Rows.ContainsKey("Best improvement")) {
          table.Rows.Add(new DataRow("Best improvement") { VisualProperties = { StartIndexZero = true } });
        }
        table.Rows["Best improvement"].Values.Add(Math.Max(cxMaxImprovement, mutMaxImprovement));
      }
      return base.Apply();
    }
  }
}
